# image_proc.py
# process (png) images to be used for Bot Ross
# generate curves such that PIC-controlled drawing bot
# can draw processed images

# after processing, sends coordinate arrays to PIC32
# to be drawn by Bot Ross

# implemented by Brett Sawka, Rohan Shah, and Sam DiPietro

# For ECE 4760, December 2019

import numpy as np
import cv2 as cv

import sys
import serial




sys.setrecursionlimit(5000) # need higher recursion limit for curve gen

#image = input("musk_profile_test.jpg")

scaler = 0.5

img = cv.imread("musk_profile_test.jpg",0) # load in image
img = cv.medianBlur(img, 5) # set blur
edges = cv.Canny(img,50,60, apertureSize = 3, L2gradient=True) # run Canny algorithm
scale_w = int(edges.shape[0]*scaler) # set scale for image
scale_h = int(edges.shape[1]*scaler)

low_res = cv.resize(edges,(scale_h, scale_w), interpolation = cv.INTER_AREA) # further scale image
idx1 = low_res.shape[0]
idx2 = low_res.shape[1]

new_edges = low_res.copy() # store processed image
for i in range(0,idx1):
    for j in range(0,idx2):
        if (i > 0 and i < idx1-1 and j > 0 and j < idx2-1 and low_res[i][j] > 50):
            # further refine each point as interpreted
            new_edges[i-1][j-1] = 255
            new_edges[i-1][j] = 255
            new_edges[i-1][j+1] = 255
            new_edges[i+1][j-1] = 255
            new_edges[i+1][j] = 255
            new_edges[i+1][j+1] = 255
            new_edges[i][j-1] = 255
            new_edges[i][j+1] = 255


machine_image = cv.resize(new_edges,(200, 200), interpolation = cv.INTER_AREA) # store processed image
for i in range(0,100):
    for j in range(0,100):
         if (machine_image[i][j] > 150): # further clarify cutoff for a clear contour
                machine_image[i][j] = 255


# curve generation
# given a 2-D array representing an image,
# find a value > cutoff
# Do a depth-first search adding each neighbor as a coordinate
# and adding each coordinate to a visiting list.
# Once there are no neighbors, return to the previous spot,
# repeating this process for all non-visited pixels
# until the image is complete

x_coords = []
y_coords = []

visited_coords = []

def curve_dfs(x, y): # x indexes the inner array, y indexes the outer array
    #print(x,y)
    if machine_image[y][x] > 150: #and (x not in visited_x[:len(visited_x)-1] and y not in visited_y[:len(visited_y)-1]):
        coords = (x,y)
        if coords not in visited_coords:
            visited_coords.append(coords)
            #visited_coords.append((x,y+1))
            if x != 199 and (x+1,y) not in visited_coords and machine_image[y][x+1] > 150:
                #print("right")
                curve_dfs(x+1, y)
                if next(reversed(visited_coords)) != (253,-53): # force this value, communicates to PIC to raise pen
                    visited_coords.append((253,-53))
            if y != 199 and (x,y+1) not in visited_coords and machine_image[y+1][x] > 150:
                #print("down")
                curve_dfs(x, y+1)
                if next(reversed(visited_coords)) != (253,-53):
                    visited_coords.append((253,-53))
            if y != 199 and x != 199 and (x+1,y+1) not in visited_coords and machine_image[y+1,x+1] > 150:
                #print("down diag")
                curve_dfs(x+1,y+1)
                if next(reversed(visited_coords)) != (253,-53):
                    visited_coords.append((253,-53))
            if x != 0 and y != 199 and (x-1,y+1) not in visited_coords and machine_image[y+1,x-1] > 150:
                curve_dfs(x-1,y+1)
                if next(reversed(visited_coords)) != (253,-53):
                    visited_coords.append((253,-53))
            
            if x != 0 and (x-1,y) not in visited_coords and machine_image[y][x-1] > 150:
                #print("left")
                curve_dfs(x-1, y)
                if next(reversed(visited_coords)) != (253,-53):
                    visited_coords.append((253,-53))
            if y != 0 and (x,y-1) not in visited_coords and machine_image[y-1][x] > 150:
             #   print("up")
                curve_dfs(x,y-1)
                if next(reversed(visited_coords)) != (253,-53):
                    visited_coords.append((253,-53))
            if x != 0 and y != 0 and (x-1,y-1) not in visited_coords and machine_image[y-1][x-1] > 150:
             #   print("up diag")
                curve_dfs(x-1,y-1)
                if next(reversed(visited_coords)) != (253,-53):
                    visited_coords.append((253,-53))
            if x != 199 and y != 0 and (x+1,y-1) not in visited_coords and machine_image[y-1,x+1] > 150:
                curve_dfs(x+1,y-1)
                if next(reversed(visited_coords)) != (253,-53):
                    visited_coords.append((253,-53))
            
                
count = 300           
for i in range(0,len(machine_image)-2):
    for j in range(0,len(machine_image[0])-2):
        if (j,i) not in visited_coords:
            curve_dfs(j,i)

            
        



for coord in visited_coords:
    x_coords.append(coord[0])
    y_coords.append(200-coord[1]) # flip y-axis so it's not upside down
#print(x_coords)
#print(y_coords)
print("Image Processing Done")
size = len(x_coords)
#for a in range(0,len(y_coords)-1):
#    y_coords[i] = 200 - y_coords[i]

#x_coords = np.array(x_coords)
#y_coords = np.array(y_coords)
#plt.plot(x_coords[x_coords != 253], y_coords[y_coords != 253])

# send to PIC32

# format size into byte array
res = 0
b = [255]

while size > 253: # format into byte-size values (excluding serial commands)
    b.append(253)
    size -= 253
b.append(size)
b.insert(0,254) # prepend start command
b.append(255) # append end command
print(b)
x_coords.insert(0,254) # prepend start command
x_coords.append(255) # append end command

y_coords.insert(0,254)
y_coords.append(255)

port_name = "COM4" # set serial port
ser = serial.Serial(port_name, 38400, timeout=10, writeTimeout=10)
ser.flushInput()
ser.flushOutput() # flush input and output
ser.write(bytearray(b)) # send the size
ser.flushOutput()

#while (ser.read(1) != str(0xFB)):
#    print("waiting")
r = 0
r = ser.read(1) # wait for acknowledgement
while r != str(0xfb): # wait for PIC to communicate it's ready
    print("waiting for size ack")
    print(r)
    r = ser.read(1)
print("after read until")
print(r)
ser.flushInput()

ser.write(bytearray(x_coords)) # send x array
ser.flushOutput()
r = ser.read(1)
while r != str(0xfb): # wait for ready
    print("waiting")
    print(r)
    r = ser.read(1)
print("after next read until")
print(r)
ser.flushInput()
ser.write(bytearray(y_coords)) # send y array

print("finished sending image") # done