#include "config_1_3_2.h"
#include "pt_cornell_1_3_2.h"
#include "tft_master.h"
#include "tft_gfx.h"


#define BUFFER_FULL  0xFF
#define BUFFER_EMPTY 0xFE
#define MAX_IMG_SIZE 5000 
static struct pt pt_UART;


/* Struct that handles the img input from Host */
typedef struct {

    unsigned char X[MAX_IMG_SIZE]; // Array of the XY-Coordinates for the Image
	unsigned char Y[MAX_IMG_SIZE];

    unsigned int  size; // Ints that hold the length of the image, 200 is 
    unsigned char valid; //Valid Bit for the Image, 1 if valid

} img;

static img imageBuffer; //A buffer of 5 images that will be used to draw


/* Thread to handle serial image input
 * Data Format: Start Command + Data {x1,y1,x2,y2...} + Terminate Command 
 */

static unsigned char          cmd[50]; // Incoming command buffer
static unsigned char       temp[5002]; // Buffer for image coordinates
static unsigned char            value;
static int                      j = 1;

static char msg[50]; // Incoming command buffer 
static int    state; // The State of the Transmission Receiver
                  
/* Thread that handles the Serial communication between host */
static PT_THREAD (protothread_UART(struct pt *pt)) {
    
    PT_BEGIN(pt);
        while(1) {

            switch(state){
                case 0: // 0: Waiting
                    // terminate serial input
                    PT_terminate_char  = 0xFF;    // Termination byte from the raspberry pi
                    PT_terminate_count =   50;    // after 50 cap out, size max 1000
                    PT_terminate_time  = 5000;    // after 5 seconds

                    
                        
                        
                    // spawn non-blocking thread to get input from raspberry pi
                    // make sure that the start command format is 0x## + '\r' so
                    // that the function properly returns
                    PT_SPAWN(pt, &pt_input, PT_GetMachineBuffer(&pt_input) );
                    if(PT_timeout==0) 
                    {
                        if((char) PT_term_buffer[0] == (char) 0xFE){
                            if ( imageBuffer.valid == 0 ) // Image buffer is empty
                            {
                                // Add up the values in the size array
                                while( (j < 50) && (PT_term_buffer[j] != (char) 0xFF))
                                {
                                    imageBuffer.size = imageBuffer.size + (unsigned int) PT_term_buffer[j]; // Get the size of the array
                                    j++;

                                }
                                // Ensure that Size > 0 so we do not immediately exit
                                if (imageBuffer.size > 0 ) {
                                    state = 1; // Move to the next state X
                                    // Send 0xFB back to host
                                    PT_send_buffer[0] = (char) 0xFB;
                                    PT_SPAWN(pt, &pt_DMA_output, PT_DMA_PutSerialBuffer(&pt_DMA_output));
                                } //Move to the next state
                            }
                            else
                            {
                                // Exit failure
                                PT_send_buffer[0] = (char) 0xFC;
                                PT_SPAWN(pt, &pt_DMA_output, PT_DMA_PutSerialBuffer(&pt_DMA_output));
                            }
                        }
                        else
                        {
                            // Exit failure
                            PT_send_buffer[0] = (char) 0xFC;
                            PT_SPAWN(pt, &pt_DMA_output, PT_DMA_PutSerialBuffer(&pt_DMA_output));
                        }
                    }
                    else 
                    {
                        // no actual command :(
                        cmd[0] = 0 ;
                        PT_send_buffer[0] = (char) 0xFC;
                        PT_SPAWN(pt, &pt_DMA_output, PT_DMA_PutSerialBuffer(&pt_DMA_output));
                    }
                    break;
                case 1: // 1: Received Size
                    // terminate serial input
                    PT_terminate_char  = 0xFF;    // Termination byte from the raspberry pi
                    PT_terminate_count = 5002;    // after 50 cap out, size max 1000
                    PT_terminate_time  = 5000;    // after 5 seconds

                    // spawn non-blocking thread to get input from raspberry pi
                    // make sure that the start command format is 0x## + '\r' so
                    // that the function properly returns
                    PT_SPAWN(pt, &pt_input, PT_GetMachineBuffer(&pt_input) );
                    if(PT_timeout==0) 
                    {
                        //sscanf( PT_term_buffer , "%u", temp);
                        if((char) PT_term_buffer[0] ==  (char) 0xFE)
                        {
                                // Move input X coordinates from serial to buffer
								int q;
								for (q = 1; q < imageBuffer.size + 1; q++) // Chop start and end off 
								{
									imageBuffer.X[q-1] = (unsigned char) PT_term_buffer[q];
								}
                                
                                state = 2; // Move to the next state (Y)

                                // Send 0xFB back to host
                                PT_send_buffer[0] =  (char) 0xFB;
								PT_SPAWN(pt, &pt_DMA_output, PT_DMA_PutSerialBuffer(&pt_DMA_output));
                        }
                        else
                        {
                            // Exit failure
                           PT_send_buffer[0] = (char) 0xFC;
                           PT_SPAWN(pt, &pt_DMA_output, PT_DMA_PutSerialBuffer(&pt_DMA_output)); 
                        }
                    }
                    else 
                    {
                        // Exit failure
                        PT_send_buffer[0] = (char) 0xFC;
                        PT_SPAWN(pt, &pt_DMA_output, PT_DMA_PutSerialBuffer(&pt_DMA_output));
                    }
                    break;
                case 2: // Get the Y coords 

                    PT_terminate_char  = 0xFF;    // Termination byte from the Host 
                    PT_terminate_count = 5002;    // after 50 cap out, size max 1000
                    PT_terminate_time  = 5000;    // after 5 seconds

                    // spawn non-blocking thread to get input from UART
                    PT_SPAWN(pt, &pt_input, PT_GetMachineBuffer(&pt_input) );
                    if(PT_timeout==0) 
                    {
                        //sscanf( PT_term_buffer , "%u", temp);
                        if((char) PT_term_buffer[0] ==  (char) 0xFE)
                        {
                                // Move input Y coordinates from serial to buffer
								int q;
								for (q = 1; q < imageBuffer.size + 1; q++) // Chop start and end off 
								{
									imageBuffer.Y[q-1] = (unsigned char) PT_term_buffer[q];
								}
                                
                                state = 3; // Full Image received move to draw state

                                // Send 0xFB back to Host
                                PT_send_buffer[0] =  (char) 0xFB;
								PT_SPAWN(pt, &pt_DMA_output, PT_DMA_PutSerialBuffer(&pt_DMA_output));
                        }
                        else
                        {
                            // Send Exit Command
                           PT_send_buffer[0] = (char) 0xFC;
                           PT_SPAWN(pt, &pt_DMA_output, PT_DMA_PutSerialBuffer(&pt_DMA_output)); 
                        }
                    }
                    else 
                    {
                        // Send Exit Command
                        PT_send_buffer[0] = (char) 0xFC;
                        PT_SPAWN(pt, &pt_DMA_output, PT_DMA_PutSerialBuffer(&pt_DMA_output));
                    }
                    break;
                case 3:
                    // SPAWN DRAW THREAD
                    imageBuffer.valid = 1;
                    state = 0; // Reset Once done
            }
        }
    PT_END(pt);
}

/* Main Funciton */
int main(void) {
    
  
	max_chars = 5002; // Set the max chars of the serial lib 
                      // to sizeof(imageBuffer) + start and end
    
    //set up interrupts system wide
    INTEnableSystemMultiVectoredInt();

    PT_setup();
    // Init threads
    PT_INIT(&pt_UART);

    while (1) {
      PT_SCHEDULE(protothread_UART(&pt_UART));
    } // Never exit while loop

  return (EXIT_SUCCESS);
}