/* 
 * File:   stepper_control.c
 * Author: Brett Sawka, Rohan Shah, Sam DiPietro
 *
 * Created on November 7, 2019, 6:20 PM
 * 
 * Contains functions for driving stepper motors on PIC32 big board
 * for ECE4760 at Cornell University
 * 
 * Pin connections to motor control signals:
 * 
 */

#include <stdio.h>
#include <stdlib.h>
#include "pt_cornell_1_3_2.h"
#include "config_1_3_2.h"
#include "port_expander_brl4.h"
//#include "tft_gfx.h"
//#include "tft_master.h"
#include <math.h>
#include "images.h"


volatile int big_counter = 0;        // count for completing one step
volatile int small_counter = 0;

volatile static int big_dir;         // indicates clockwise (0) or counter clockwise (1)
volatile static int small_dir;

volatile int big_step_in_progress;   // 1 when step is in progress
volatile int small_step_in_progress;

volatile static int is_big_step = 0;
volatile static int is_small_step = 0;

static int x; // Xpos
static int y; // Ypos

static char buffer[30];

/* Go through a cycle towards stepping the motor, runs at 83 Hz */
void __ISR(_TIMER_2_VECTOR, ipl2) Timer2Handler(void) {
    
    mT2ClearIntFlag();
    
    if (is_big_step) {
        // increment or 0 counter
        //counter++;
        if (big_counter > 3) big_counter = 0;
        if (big_dir) { // counter clockwise
                switch (big_counter) {

                    case 0: 
                        setBits(GPIOZ, BIT_0 | BIT_2);
                        clearBits(GPIOZ, BIT_1 | BIT_3);
                        big_step_in_progress = 0;
                        is_big_step = 0;
                        break;
                    case 1:  

                        setBits(GPIOZ, BIT_1 | BIT_2);
                        clearBits(GPIOZ, BIT_0 | BIT_3);
                        big_step_in_progress = 0;
                        is_big_step = 0;
                        break;
                    case 2:  

                        setBits(GPIOZ, BIT_1 | BIT_3);
                        clearBits(GPIOZ, BIT_0 | BIT_2);
                        big_step_in_progress = 0;
                        is_big_step = 0;
                        break;
                    case 3:

                        setBits(GPIOZ, BIT_0 | BIT_3);
                        clearBits(GPIOZ, BIT_1 | BIT_2);
                        big_step_in_progress = 0;
                        is_big_step = 0;
                        
                        break;
                    default:
                        clearBits(GPIOZ, BIT_0 | BIT_1 | BIT_2 | BIT_3);
                }
            }
        else { // clockwise
                switch (big_counter) {

                    case 0: 

                        setBits(GPIOZ, BIT_0 | BIT_2);
                        clearBits(GPIOZ, BIT_1 | BIT_3);
                        big_step_in_progress = 0;
                        is_big_step = 0;

                        break;
                    case 1: 

                        setBits(GPIOZ, BIT_0 | BIT_3);
                        clearBits(GPIOZ, BIT_1 | BIT_2);
                        big_step_in_progress = 0;
                        is_big_step = 0;

                        break;
                    case 2: 

                        setBits(GPIOZ, BIT_1 | BIT_3);
                        clearBits(GPIOZ, BIT_0 | BIT_2);
                        big_step_in_progress = 0;
                        is_big_step = 0;

                        break;
                    case 3:   

                        setBits(GPIOZ, BIT_1 | BIT_2);
                        clearBits(GPIOZ, BIT_0 | BIT_3);
                        big_step_in_progress = 0;
                        is_big_step = 0;
                        
                        break;
                    default:
                        clearBits(GPIOZ, BIT_0 | BIT_1 | BIT_2 | BIT_3);

                }   
    
    
            }
        big_counter++;
    }
    if (is_small_step) {
        // increment or 0 counter
        //counter++;
        if (small_counter > 3) small_counter = 0;
        if (small_dir) {
                switch (small_counter) {

                    case 0: 
                        setBits(GPIOY, BIT_0 | BIT_2);
                        clearBits(GPIOY, BIT_1 | BIT_3);
                        small_step_in_progress = 0;
                        is_small_step = 0;
                        break;
                    case 1:  

                        setBits(GPIOY, BIT_1 | BIT_2);
                        clearBits(GPIOY, BIT_0 | BIT_3);
                        small_step_in_progress = 0;
                        is_small_step = 0;
                        break;
                    case 2:  

                        setBits(GPIOY, BIT_1 | BIT_3);
                        clearBits(GPIOY, BIT_0 | BIT_2);
                        small_step_in_progress = 0;
                        is_small_step = 0;
                        break;
                    case 3:

                        setBits(GPIOY, BIT_0 | BIT_3);
                        clearBits(GPIOY, BIT_1 | BIT_2);
                        small_step_in_progress = 0;
                        is_small_step = 0;
                        
                        break;
                    default:
                        clearBits(GPIOY, BIT_0 | BIT_1 | BIT_2 | BIT_3);

                }
            }
            else {
                switch (small_counter) {

                    case 0: 

                        setBits(GPIOY, BIT_0 | BIT_2);
                        clearBits(GPIOY, BIT_1 | BIT_3);
                        small_step_in_progress = 0;
                        is_small_step = 0;

                        break;
                    case 1: 

                        setBits(GPIOY, BIT_0 | BIT_3);
                        clearBits(GPIOY, BIT_1 | BIT_2);
                        small_step_in_progress = 0;
                        is_small_step = 0;

                        break;
                    case 2: 

                        setBits(GPIOY, BIT_1 | BIT_3);
                        clearBits(GPIOY, BIT_0 | BIT_2);
                        small_step_in_progress = 0;
                        is_small_step = 0;

                        break;
                    case 3:   

                        setBits(GPIOY, BIT_1 | BIT_2);
                        clearBits(GPIOY, BIT_0 | BIT_3);
                        small_step_in_progress = 0;
                        is_small_step = 0;
                        
                        break;
                    default:
                        clearBits(GPIOY, BIT_0 | BIT_1 | BIT_2 | BIT_3);
                }   
            }
        small_counter++;
    }
}



#define pen_down() (SetDCOC1PWM(4290)) // set pwm duty cycle such that the pen servo points down
#define pen_up()   (SetDCOC1PWM(3500)) // set pwm duty cycle such that pen servo points horizontal (bringing pen up)

static struct pt pt_draw, pt_big_control, pt_small_control, pt_big_step, pt_small_step; //pt_display;

static int turn;
static int x_steps = 0;
static int is_pen_down = 0;

// parameters for image drawing
// Specifically for flash import
#define IMAGE_SIZE 3229 // mickey - 862, yoshi - 3477
#define PEN_INSTR   253





static int i; // index of image arrays
//static int dx; // x difference
//static int dy; // y difference
static int sx; // up or down direction depending on x coords
static int sy; // up or down direction depending on y coords
//static int err; // error based on both x and y
//static int err2; // 2*err
static int x1; // x_coord[i] + 1 (next x position)
static int y1; // y_coord[i] + 1 (next y position)
static int x0; // x_coord[i] (current x pos)
static int y0; // y_coord[i] (current y pos)
static int move = 1;

/* Linear Interpolation Algorithm */
static PT_THREAD (protothread_draw(struct pt *pt)) {
    
    PT_BEGIN(pt);
    
    
    
    pen_up(); // raise the pen to go to starting position
    
    for (i = 0; i < IMAGE_SIZE-1; i++) {
        
        
        // set all variables for this transition
        
        if (image_x[i+1] == PEN_INSTR && image_y[i+1] == PEN_INSTR) { // got instruction to lift pen
            pen_up();
            is_pen_down = 0;
            x0 = image_x[i]; // save this coordinate as next x0
            y0 = image_y[i]; // save this coordinate as next y0
            move = 0; // don't move to any position
        }
        else if (image_x[i] == PEN_INSTR && image_y[i] == PEN_INSTR) {
            pen_up();
            is_pen_down = 0;
            x1 = image_x[i+1]; // use next coordinate as x1
            y1 = image_y[i+1]; // use next coordinate as y1
            move = 1; // move (with pen up)
        }
        else {
            // set current and target coordinates
            x0 = image_x[i];
            y0 = image_y[i];
            x1 = image_x[i+1];
            y1 = image_y[i+1];
            move = 1; // move (with pen down)
        }
        tft_fillRoundRect(0,110, 100, 14, 1, ILI9340_BLACK);// x,y,w,h,radius,color
        tft_setCursor(0, 110);
        tft_setTextColor(ILI9340_YELLOW); tft_setTextSize(1);
        sprintf(buffer,"(move: %d", move);
        tft_writeString(buffer);
        //dx = abs(x1 - x0);
        sx = x0 < x1 ? 1 : -1; // set sign of difference between x
        //dy = -abs(y1 - y0);
        sy = y0 < y1 ? 1: -1; // set sign of difference between y
        //err = dx+dy;
        
        // print out location it's headed to (assuming move == 1)

        // move stepper one unit in the correct direction, then update all variables.
        // repeat until x0 == x1 and y0 == y1, then go to the next point
        /*if (image_x[i] == PEN_INSTR && image_y[i] == PEN_INSTR) {
            pen_up(); // raise the pen when the forced instruction value of 300 is seen
        }*/

        if (move) { // only if we're supposed to move
            while (x0 != x1 || y0 != y1) {
                // move steppers in correct direction (if values are still unequal)
                if (x0 != x1) {

                    if (sx == 1) { // move motor in +x direction
                        PT_SPAWN(pt, &pt_small_step, small_step(&pt_small_step, 1, 1) );
                        x0 += sx;
                    }
                    else { // move motor in -x direction
                        PT_SPAWN(pt, &pt_small_step, small_step(&pt_small_step, 1, 0) );
                        x0 += sx;
                        if (is_pen_down) {
                            x_steps++; //Increment the adjustment counter
                        }
                    }
                    
                    // If we have stepped 30 times in the -x direction w/ pen down, step one more   
                    if (x_steps >= 30) { 
                    	// Make the adjustment
                        PT_SPAWN(pt, &pt_small_step, small_step(&pt_small_step, 1, 0) );
                        x_steps = 0; // Reset the counters
                    }
                }

                // Check the directionality of the y step
                if (y0 != y1) {
                    if (sy == 1) {
                        PT_SPAWN(pt, &pt_big_step, big_step(&pt_big_step, 1, 1));
                        y0 += sy;
                    }
                    else {
                        PT_SPAWN(pt, &pt_big_step, big_step(&pt_big_step, 1, 0));
                        y0 += sy;
                    }
                }  
            }
        pen_down(); // assure the pen is lowered after a coordinate is traveled to
        is_pen_down = 1;
        }
    } 
    pen_up(); // raise the pen, we're done
    mPORTASetBits(BIT_0); // turn on board LED to signify done
    while(1) {
        PT_YIELD(pt);
    }
    PT_END(pt);
}


/* Used for debugging, causes the motors to free spin


// Thread that is spawned to turn the big motor (y) one time 
static PT_THREAD (protothread_big_control(struct pt *pt)) {

    PT_BEGIN(pt);
    
        while(1) {
            if (1) {
                //pen_down();
                PT_SPAWN(pt, &pt_big_step, big_step(&pt_big_step, 1, 0) );
                turn = 0;
            }
            else {
                PT_YIELD(pt);
            }
        }
    
    PT_END(pt);

}

// Thread that is spawned to turn the small motor (x) one time
static PT_THREAD (protothread_small_control(struct pt *pt)) {

    PT_BEGIN(pt);
    static int c = 0;
    turn = 0;

        while(1) {
            
            if (!turn) {
                
                PT_SPAWN(pt, &pt_small_step, small_step(&pt_small_step, 1, 0) );
            }
            else {
                PT_YIELD(pt);
            }

        }
    
    PT_END(pt);
}

*/

/* Thread that actually generates the output signal for small motor */
int big_step(struct pt *pt, int num_big_steps, int dir) {

    static int i;
    PT_BEGIN(pt);
        i = 0;
        big_dir = dir;
        while (i < 27*num_big_steps) {
            is_big_step = 1;
            PT_YIELD_UNTIL(pt, !is_big_step);
            //mPORTAToggleBits(BIT_0);
            y++;
            i++;
        }
        
        clearBits(GPIOZ, BIT_0 | BIT_1 | BIT_2 | BIT_3);
        //mPORTAToggleBits(BIT_0);
        PT_EXIT(pt);
        
    PT_END(pt);

}

/* Thread that actually generates the output signal for small motor */
int small_step(struct pt *pt, int num_small_steps, int dir) {
    
    static int j;
    PT_BEGIN(pt);
        //mPORTAToggleBits(BIT_0);
        j = 0;
        small_dir = dir;
        while (j < 50*num_small_steps) {
            is_small_step = 1;
            PT_YIELD_UNTIL(pt, !is_small_step);
            x++;
            j++;

        }
        
        clearBits(GPIOY, BIT_0 | BIT_1 | BIT_2 | BIT_3);
        //mPORTAToggleBits(BIT_0);
        PT_EXIT(pt);
    
    PT_END(pt);
    
}


/* Main function */
int main(void) {
    
    ANSELA= 0; ANSELB = 0;
    
    // init pins used for stepper motor control, setup timer 2, enable interrupts
   
    initPE();
    
    INTEnableSystemMultiVectoredInt(); // enable interrupts
    mPortZSetPinsOut(BIT_0 | BIT_1 | BIT_2 | BIT_3);
    mPortYSetPinsOut(BIT_0 | BIT_1 | BIT_2 | BIT_3);
    
    mPORTASetPinsDigitalOut(BIT_0);
    
    
    clearBits(GPIOZ, BIT_0 | BIT_1 | BIT_2 | BIT_3); // BIT_0 and BIT_1 are A+ and A-
                                                     // BIT_2 and BIT_3 are B+ and B-
    clearBits(GPIOY, BIT_0 | BIT_1 | BIT_2 | BIT_3);
    
     // set up timer2
    OpenTimer2(T2_ON | T2_SOURCE_INT | T2_PS_1_8, 100000); // runs at roughly 83 Hz
    // Need ISR to step motors
    ConfigIntTimer2(T2_INT_ON | T2_INT_PRIOR_2);
    mT2ClearIntFlag(); // and clear the interrupt flag
    
    // open timer 3 for output compare
    OpenTimer3(T3_ON | T3_SOURCE_INT | T3_PS_1_16, 50000);
    // open output compare channel for pen servo pwm
    OpenOC1(OC_ON | OC_TIMER3_SRC | OC_PWM_FAULT_PIN_DISABLE , 0, 0);
    // map to RPB7
    PPSOutput(1, RPB7, OC1);
    pen_up();
   
    
    PT_setup();
    
    PT_INIT(&pt_big_control);
    PT_INIT(&pt_small_control);
    //PT_INIT(&pt_display);
    
    PT_INIT(&pt_draw);
    
    
	    // example usage - rotate the motor clockwise 200 steps (360 degrees)
	    /* int i = 0;
	    big_dir = 0;
	    small_dir = 0;
	    while (i < 400) {
	        big_step_in_progress = 1; // notify step in progress
	        small_step_in_progress = 1;
	        is_big_step = 1; // tell ISR to run a step
	        is_small_step = 1;
	        while(big_step_in_progress && small_step_in_progress); // wait until step is done
	        i++;
	    }
	  	*/   
	    //clearBits(GPIOZ, BIT_0 | BIT_1 | BIT_2 | BIT_3);
	    //clearBits(GPIOY, BIT_0 | BIT_1 | BIT_2 | BIT_3);

    mPORTAClearBits(BIT_0);
    

    
    // step control
    while (1) {
             
        PT_SCHEDULE(protothread_draw(&pt_draw));
        
        //PT_SCHEDULE(protothread_big_control(&pt_big_control));
        //PT_SCHEDULE(protothread_small_control(&pt_small_control));
        //PT_SCHEDULE(protothread_display(&pt_display));
        //PT_SCHEDULE(protothread_idle(&pt_idle));

        
    } // never exit while!
    return (EXIT_SUCCESS);
}

